#include <cassert>
#include <iostream>
#include <cmath>
#include "cloth.hxx"

const float Cloth::k = 200;
const float Cloth::gamma = 25;
//const float Cloth::dt = 5e-4;
float Cloth::dt = 1e-4;
static float h = 0.0;

static SbVec3f d_force(const SbVec3f &l, const SbVec3f &r, const float l0) {
  SbVec3f d(l - r); /* f(l - r) = -f(r - l) ! */
  float d_length = d.length();

  assert(Cloth::k > 0);
  assert(l0 != 0.0);
  assert(d_length > 0);
  if (fabs(d_length - h)/h < 0.01) return SbVec3f(0,0,0);
  return Cloth::k*(l0/d_length - 1)*d;
}

SbVec3f Cloth::compute_force(const SbVec3f vertices[], const SbVec3f velocities[], size_t idx) const {
  SbVec3f force(0,0,0);

  if (this ->is_fixed(idx)) return force;
  assert(force[0] == 0 && force[1] == 0 && force[2] == 0);

  if (idx%this -> x_max() > 0) {
    force += d_force(vertices[idx], vertices[idx-1], h);
    if (idx/this -> x_max() > 0) force += d_force(vertices[idx], vertices[idx-this -> x_max()-1], M_SQRT2*h);
    if (idx/this -> x_max() < this -> y_max()-1) force += d_force(vertices[idx], vertices[idx+this -> x_max()-1], M_SQRT2*h);
  }

  if (idx%this -> x_max() < this -> x_max()-1) {
    force += d_force(vertices[idx], vertices[idx+1], h);
    if (idx/this -> x_max() > 0) force += d_force(vertices[idx], vertices[idx-this -> x_max()+1], M_SQRT2*h);
    if (idx/this -> x_max() < this -> y_max()-1) force += d_force(vertices[idx], vertices[idx+this -> x_max()+1], M_SQRT2*h);
  }

  if (idx/this -> x_max() > 0) force += d_force(vertices[idx], vertices[idx-this -> x_max()], h);
  if (idx/this -> x_max() < this -> y_max()-1) force += d_force(vertices[idx], vertices[idx+this -> x_max()], h);

  force -= Cloth::gamma*velocities[idx];
  force += ext_forces[idx];
  force[1] -= 9.81;

  assert(!isnan(force[0]) && !isnan(force[1]) && !isnan(force[2]));
  return force;
}

static void compute_forces(SbVec3f dv[], const SbVec3f v[], const SbVec3f x[], const Cloth &cloth) {
  for (size_t i = 0; i < cloth.n(); i++) dv[i] = cloth.compute_force(x,v,i);
}

/* performs a simple integration using given R-K-Scheme */
class RK_integrator {
  Cloth &cloth;
  const size_t n;

  SbVec3f *ks[2]; /* derivatives, 0 -> x, 1 -> v */
  SbVec3f *zs[2]; /* intermediate approximations, 0 ->  */

  /*  the RK-scheme */
  const size_t s;
  const float *a, *c;

  static const float int_tol;
  static const float max_dt; 
  
public:
  /* 
     format of scheme:
     a: linear array with size (s+1)*(s+1), holding the coefficients for the approximate derivates (1 <= i <= (s+1))
     c: linear array with size (s+1), holding the coefficients for the "t" variable in integration (1 <= i <= s+1)
  */
  RK_integrator(Cloth &cloth, size_t s, const float a[], const float c[])
    : cloth(cloth), n(cloth.n()), s(s+1), a(a), c(c) {
    ks[0] = new SbVec3f[this->s*n];
    ks[1] = new SbVec3f[this->s*n];
    zs[0] = new SbVec3f[(this->s+1)*n];
    zs[1] = new SbVec3f[(this->s+1)*n];
  }

  ~RK_integrator(void) {
    assert(ks[0] != NULL);
    assert(ks[1] != NULL);
    assert(zs[0] != NULL);
    assert(zs[1] != NULL);

    delete ks[0];
    delete ks[1];
    delete zs[0];
    delete zs[1];
  }
  
  void integrate(SbVec3f x1[], const SbVec3f x0[], 
		 SbVec3f v1[], const SbVec3f v0[],
		 float &dt) {
    /*
      Be aware that "s" is larger than the scheme's true "s", by one!
    */
    assert(cloth.n() == n);
    assert(dt > 0);
    assert(c[s-1] == 1);

    SbVec3f *intermediate_k[2];    


    SbVec3f *low_ord_appro[2];
    
    low_ord_appro[0] = new SbVec3f[n];
    low_ord_appro[1] = new SbVec3f[n];

    intermediate_k[0] = new SbVec3f[n];
    intermediate_k[1] = new SbVec3f[n];
    
    memcpy(zs[0], x0, n*sizeof(SbVec3f));
    memcpy(zs[1], v0, n*sizeof(SbVec3f));

    for (size_t i = 1; i < s+1; i++) {
      float tdt = dt*c[i-1];		

      /* "compute" derivatives */
      compute_forces(&ks[1][(i-1)*n], &zs[1][(i-1)*n], &zs[0][(i-1)*n], cloth);
      memcpy(&ks[0][(i-1)*n], &zs[1][(i-1)*n], n*sizeof(SbVec3f));

      /*
	Here comes to clue with this algorithm, it makes 
	timesteps of variable length, algorithm as explained in winter term
	"wissenschaftliches Rechnen" 04, based on Matlab's Runge-Kutta 4(5):
	Depending on the difference of the fourth (s-1-th) order approximation and fifth (s-th)
	order approximation the timesteps is either decreased (defference great) or decreased 
	(difference low), thus the time step is depending on the turbulence of the force function
	at current time.

	appro = 0, (s-1)-th order approximation
	appro = 1, s-th order approximation
      */

      {
	int appro = 0;

      recompute_final:	
	/* compute intermediate approximation i */		
	for (int z = 0; z < 2; z++) {
	  /* nullify buffer */
	  for (int v = n-1; v >= 0; v--) intermediate_k[z][v].setValue(0,0,0);
	  
	  for (size_t k = 0; k < i; k++) {
	      float derivative_coeff = a[(i-1+appro)*s+k];

	      for (size_t v = 0; v < n; v++) {
		intermediate_k[z][v] += derivative_coeff*ks[z][k*n+v];
		assert(k*n+v < s*n);
	      } /* for (v) */
	    } 
	  } /* sum a*k */

	  for(int z = 0; z < 2; z++) {	
	    for (size_t v = 0; v < n; v++) {
	      zs[z][i*n+v] = zs[z][v] + tdt*intermediate_k[z][v];	    
	      assert(i*n+v < n*(s+1));
	    } /* for (v) */
	  } 
	  /* end of inner approximation step */	          

	  /* save (s-1)-th order approximation for comparison */
	  if (i == s) {
	      if (appro == 0) {
		memcpy(low_ord_appro[0], &zs[0][s*n], n*sizeof(SbVec3f));
		memcpy(low_ord_appro[1], &zs[1][s*n], n*sizeof(SbVec3f));
		appro = 1;
		goto recompute_final;
	      } else {
		/* compute dt used in next timestep */
		float avg_diff = 0, avg_better = 0;
		float odt = dt;
	    
		for (int v = 0; v < (int)n; v++) {
		  avg_diff += (low_ord_appro[0][v] - zs[0][s*n+v]).length();
		  avg_better += zs[0][s*n+v].length();
		}

		avg_diff /= (float)n;	
		avg_better /= (float)n;	
		dt = dt*std::pow(int_tol*avg_better/avg_diff,(float)1.0/5); 		
		if (dt > max_dt) dt = max_dt;		
		if (odt > 2*dt) dt = 1.01*odt; 
	      }
	  }
      }
    }/* for (i), "scheme level loop" */    
    
  
    /* copy last approximation as final one */
    memcpy(x1, &zs[0][s*n], sizeof(SbVec3f)*n);
    memcpy(v1, &zs[1][s*n], sizeof(SbVec3f)*n);

    delete low_ord_appro[0];
    delete low_ord_appro[1];

    delete intermediate_k[0];
    delete intermediate_k[1];    
  } /* integrate() */
}; /* RK_integrator */
const float RK_integrator::int_tol = 1e-3;
const float RK_integrator::max_dt = 5e-2; /* 20 frames / sec */

void swap_buffers(SbVec3f* &r, SbVec3f* &l) {
  SbVec3f *tmp;

  tmp = r;
  r = l;
  l = tmp;
}

// void Cloth::update() {
//   const size_t n = xMax*y_max();

//   for (size_t v = 0; v < n; v++) {
//     if (!fixed[v]) next_velocities[v] = velocities[v] + dt*compute_force(vertices, velocities, v);
//   }

//   for (size_t v = 0; v < n; v++) {
//     if (!fixed[v]) next_vertices[v] = vertices[v] + (float)(dt*0.5)*(velocities[v] + next_velocities[v]);    
//     assert((next_vertices[v] - vertices[v]).length() <= 1);
//     assert(!fixed[v] || next_vertices[v] == vertices[v]);
//   }
//   swap_buffers(velocities, next_velocities);
//   swap_buffers(vertices, next_vertices);
  
//   for (size_t v = 0; v < n; v++) ext_forces[v].setValue(0,0,0);    
// }

void Cloth::update() {
  float a[] = 
    {1.0/5, 0, 0, 0, 0, 0, 0,
     3.0/40, 9.0/40, 0, 0, 0, 0, 0,
     44.0/45,-56.0/15,32.0/9, 0, 0, 0, 0,
     19372.0/6561, -25360.0/2187, 64448.0/6561, -212.0/729, 0, 0, 0,
     9017.0/3168, -355.0/33, 46731.0/5247, 49.0/179, -5103.0/18656, 0, 0,
     35.0/384, 0, 500.0/1113, 125.0/192, -2187.0/6784, 11.0/84, 0,
     35.0/384, 0, 500.0/1113, 125.0/192, -2187.0/6784, 11.0/84, 0,
     5179.0/57600, 0, 7571.0/16695, 393.0/640, -92097.0/339200, 187.0/2100, 1.0/40
    };

  float c[] = {1.0/5, 3.0/10, 4.0/5, 8.0/9, 1, 1, 1};
  RK_integrator integrator(*this, 6, a, c);
  
  integrator.integrate(next_vertices, vertices, next_velocities, velocities, dt);

  swap_buffers(velocities, next_velocities);
  swap_buffers(vertices, next_vertices);

  for (size_t v = 0; v < n(); v++) this -> ext_forces[v].setValue(0,0,0);    
}

void Cloth::fix_masspoint(size_t x, size_t y, const SbVec3f &pos) {
  size_t idx = y*xMax+x;

  assert(y < yMax && x < xMax);  
  fixed[idx] = true;
  vertices[idx] = pos;
  next_vertices[idx] = pos;
  next_velocities[idx] = SbVec3f(0,0,0);
  velocities[idx] = SbVec3f(0,0,0);
}

Cloth::Cloth() : xMax(10), yMax(10), cloth_size(yMax*xMax) {
  const size_t n = xMax*yMax;

  velocities = new SbVec3f[n];
  next_velocities = new SbVec3f[n];
  vertices = new SbVec3f[n];
  next_vertices = new SbVec3f[n];
  fixed = new bool[n];
  ext_forces = new SbVec3f[n];

  for (size_t v = 0; v < n; v++) {
    velocities[v].setValue(0,0,0);
    ext_forces[v].setValue(0,0,0);
    fixed[v] = false;
  }

  {
    SbVec3f pt(0,0,0);

    h = 5.0/(xMax-1);
    for (size_t v = 0; v < n; v++) {      
      if (v != 0 && v % xMax == 0) {
	pt[1] += h;      
	pt[0] = 0;
      } 
      vertices[v] = pt;
      pt[0] += h;
    }
  } /* vertex init */    

  assert(h > 0);
  assert(k > 0);
  assert(dt > 0);
}
